CHIPSET(0x1602, bdw_gt1, "BDW GT1", "Intel(R) HD Graphics")
CHIPSET(0x1606, bdw_gt1, "BDW GT1", "Intel(R) HD Graphics")
CHIPSET(0x160a, bdw_gt1, "BDW GT1", "Intel(R) HD Graphics")
CHIPSET(0x160b, bdw_gt1, "BDW GT1", "Intel(R) HD Graphics")
CHIPSET(0x160d, bdw_gt1, "BDW GT1", "Intel(R) HD Graphics")
CHIPSET(0x160e, bdw_gt1, "BDW GT1", "Intel(R) HD Graphics")
CHIPSET(0x1612, bdw_gt2, "BDW GT2", "Intel(R) HD Graphics 5600")
CHIPSET(0x1616, bdw_gt2, "BDW GT2", "Intel(R) HD Graphics 5500")
CHIPSET(0x161a, bdw_gt2, "BDW GT2", "Intel(R) HD Graphics P5700")
CHIPSET(0x161b, bdw_gt2, "BDW GT2", "Intel(R) HD Graphics")
CHIPSET(0x161d, bdw_gt2, "BDW GT2", "Intel(R) HD Graphics")
CHIPSET(0x161e, bdw_gt2, "BDW GT2", "Intel(R) HD Graphics 5300")
CHIPSET(0x1622, bdw_gt3, "BDW GT3", "Intel(R) Iris(R) Pro Graphics 6200")
CHIPSET(0x1626, bdw_gt3, "BDW GT3", "Intel(R) HD Graphics 6000")
CHIPSET(0x162a, bdw_gt3, "BDW GT3", "Intel(R) Iris(R) Pro Graphics P6300")
CHIPSET(0x162b, bdw_gt3, "BDW GT3", "Intel(R) Iris(R) Graphics 6100")
CHIPSET(0x162d, bdw_gt3, "BDW GT3", "Intel(R) HD Graphics")
CHIPSET(0x162e, bdw_gt3, "BDW GT3", "Intel(R) HD Graphics")

CHIPSET(0x1902, skl_gt1, "SKL GT1",  "Intel(R) HD Graphics 510")
CHIPSET(0x1906, skl_gt1, "SKL GT1",  "Intel(R) HD Graphics 510")
CHIPSET(0x190a, skl_gt1, "SKL GT1",  "Intel(R) HD Graphics")
CHIPSET(0x190b, skl_gt1, "SKL GT1",  "Intel(R) HD Graphics 510")
CHIPSET(0x190e, skl_gt1, "SKL GT1",  "Intel(R) HD Graphics")
CHIPSET(0x1912, skl_gt2, "SKL GT2",  "Intel(R) HD Graphics 530")
CHIPSET(0x1913, skl_gt2, "SKL GT2F", "Intel(R) HD Graphics")
CHIPSET(0x1915, skl_gt2, "SKL GT2F", "Intel(R) HD Graphics")
CHIPSET(0x1916, skl_gt2, "SKL GT2",  "Intel(R) HD Graphics 520")
CHIPSET(0x1917, skl_gt2, "SKL GT2F", "Intel(R) HD Graphics")
CHIPSET(0x191a, skl_gt2, "SKL GT2",  "Intel(R) HD Graphics")
CHIPSET(0x191b, skl_gt2, "SKL GT2",  "Intel(R) HD Graphics 530")
CHIPSET(0x191d, skl_gt2, "SKL GT2",  "Intel(R) HD Graphics P530")
CHIPSET(0x191e, skl_gt2, "SKL GT2",  "Intel(R) HD Graphics 515")
CHIPSET(0x1921, skl_gt2, "SKL GT2",  "Intel(R) HD Graphics 520")
CHIPSET(0x1923, skl_gt3, "SKL GT3",  "Intel(R) HD Graphics 535")
CHIPSET(0x1926, skl_gt3, "SKL GT3",  "Intel(R) Iris(R) Graphics 540")
CHIPSET(0x1927, skl_gt3, "SKL GT3",  "Intel(R) Iris(R) Graphics 550")
CHIPSET(0x192a, skl_gt4, "SKL GT4",  "Intel(R) HD Graphics")
CHIPSET(0x192b, skl_gt3, "SKL GT3",  "Intel(R) Iris(R) Graphics 555")
CHIPSET(0x192d, skl_gt3, "SKL GT3",  "Intel(R) Iris(R) Graphics P555")
CHIPSET(0x1932, skl_gt4, "SKL GT4",  "Intel(R) Iris(R) Pro Graphics 580")
CHIPSET(0x193a, skl_gt4, "SKL GT4",  "Intel(R) Iris(R) Pro Graphics P580")
CHIPSET(0x193b, skl_gt4, "SKL GT4",  "Intel(R) Iris(R) Pro Graphics 580")
CHIPSET(0x193d, skl_gt4, "SKL GT4",  "Intel(R) Iris(R) Pro Graphics P580")

CHIPSET(0x0a84, bxt,     "BXT 3",    "Intel(R) HD Graphics")
CHIPSET(0x1a84, bxt,     "BXT 3",    "Intel(R) HD Graphics")
CHIPSET(0x1a85, bxt_2x6, "BXT 2",    "Intel(R) HD Graphics")
CHIPSET(0x5a84, bxt,     "APL 3",    "Intel(R) HD Graphics 505")
CHIPSET(0x5a85, bxt_2x6, "APL 2",    "Intel(R) HD Graphics 500")

CHIPSET(0x3184, glk,     "GLK 3",    "Intel(R) UHD Graphics 605")
CHIPSET(0x3185, glk_2x6, "GLK 2",    "Intel(R) UHD Graphics 600")

CHIPSET(0x5902, kbl_gt1,   "KBL GT1",   "Intel(R) HD Graphics 610")
CHIPSET(0x5906, kbl_gt1,   "KBL GT1",   "Intel(R) HD Graphics 610")
CHIPSET(0x590a, kbl_gt1,   "KBL GT1",   "Intel(R) HD Graphics")
CHIPSET(0x5908, kbl_gt1,   "KBL GT1",   "Intel(R) HD Graphics")
CHIPSET(0x590b, kbl_gt1,   "KBL GT1",   "Intel(R) HD Graphics 610")
CHIPSET(0x590e, kbl_gt1,   "KBL GT1",   "Intel(R) HD Graphics")
CHIPSET(0x5913, kbl_gt1_5, "KBL GT1.5", "Intel(R) HD Graphics")
CHIPSET(0x5915, kbl_gt1_5, "KBL GT1.5", "Intel(R) HD Graphics")
CHIPSET(0x5917, kbl_gt2,   "KBL GT2",   "Intel(R) UHD Graphics 620")
CHIPSET(0x5912, kbl_gt2,   "KBL GT2",   "Intel(R) HD Graphics 630")
CHIPSET(0x5916, kbl_gt2,   "KBL GT2",   "Intel(R) HD Graphics 620")
CHIPSET(0x591a, kbl_gt2,   "KBL GT2",   "Intel(R) HD Graphics P630")
CHIPSET(0x591b, kbl_gt2,   "KBL GT2",   "Intel(R) HD Graphics 630")
CHIPSET(0x591d, kbl_gt2,   "KBL GT2",   "Intel(R) HD Graphics P630")
CHIPSET(0x591e, kbl_gt2,   "KBL GT2",   "Intel(R) HD Graphics 615")
CHIPSET(0x5921, kbl_gt2,   "KBL GT2F",  "Intel(R) HD Graphics 620")
CHIPSET(0x5923, kbl_gt3,   "KBL GT3",   "Intel(R) HD Graphics 635")
CHIPSET(0x5926, kbl_gt3,   "KBL GT3",   "Intel(R) Iris(R) Plus Graphics 640 (Kaby Lake GT3e)")
CHIPSET(0x5927, kbl_gt3,   "KBL GT3",   "Intel(R) Iris(R) Plus Graphics 650 (Kaby Lake GT3e)")
CHIPSET(0x593b, kbl_gt4,   "KBL GT4",   "Intel(R) HD Graphics")

CHIPSET(0x591c, kbl_gt2, "AML-KBL", "Intel(R) UHD Graphics 615")
CHIPSET(0x87c0, kbl_gt2, "AML-KBL", "Intel(R) UHD Graphics 617")

CHIPSET(0x87ca, cfl_gt2, "AML-CFL", "Intel(R) UHD Graphics")

CHIPSET(0x3e90, cfl_gt1, "CFL GT1", "Intel(R) UHD Graphics 610")
CHIPSET(0x3e93, cfl_gt1, "CFL GT1", "Intel(R) UHD Graphics 610")
CHIPSET(0x3e99, cfl_gt1, "CFL GT1", "Intel(R) UHD Graphics 610")
CHIPSET(0x3e9c, cfl_gt1, "CFL GT1", "Intel(R) UHD Graphics 610")
CHIPSET(0x3e91, cfl_gt2, "CFL GT2", "Intel(R) UHD Graphics 630")
CHIPSET(0x3e92, cfl_gt2, "CFL GT2", "Intel(R) UHD Graphics 630")
CHIPSET(0x3e96, cfl_gt2, "CFL GT2", "Intel(R) UHD Graphics P630")
CHIPSET(0x3e98, cfl_gt2, "CFL GT2", "Intel(R) UHD Graphics 630")
CHIPSET(0x3e9a, cfl_gt2, "CFL GT2", "Intel(R) UHD Graphics P630")
CHIPSET(0x3e9b, cfl_gt2, "CFL GT2", "Intel(R) UHD Graphics 630")
CHIPSET(0x3e94, cfl_gt2, "CFL GT2", "Intel(R) UHD Graphics P630")
CHIPSET(0x3ea9, cfl_gt2, "CFL GT2", "Intel(R) UHD Graphics 620")
CHIPSET(0x3ea5, cfl_gt3, "CFL GT3", "Intel(R) Iris(R) Plus Graphics 655")
CHIPSET(0x3ea6, cfl_gt3, "CFL GT3", "Intel(R) Iris(R) Plus Graphics 645")
CHIPSET(0x3ea7, cfl_gt3, "CFL GT3", "Intel(R) HD Graphics")
CHIPSET(0x3ea8, cfl_gt3, "CFL GT3", "Intel(R) Iris(R) Plus Graphics 655")

CHIPSET(0x3ea1, cfl_gt1, "WHL GT1", "Intel(R) UHD Graphics 610")
CHIPSET(0x3ea4, cfl_gt1, "WHL GT1", "Intel(R) UHD Graphics")
CHIPSET(0x3ea0, cfl_gt2, "WHL GT2", "Intel(R) UHD Graphics 620")
CHIPSET(0x3ea3, cfl_gt2, "WHL GT2", "Intel(R) UHD Graphics")
CHIPSET(0x3ea2, cfl_gt3, "WHL GT3", "Intel(R) UHD Graphics")

CHIPSET(0x9b21, cfl_gt1, "CML GT1", "Intel(R) UHD Graphics")
CHIPSET(0x9ba0, cfl_gt1, "CML GT1", "Intel(R) UHD Graphics")
CHIPSET(0x9ba2, cfl_gt1, "CML GT1", "Intel(R) UHD Graphics")
CHIPSET(0x9ba4, cfl_gt1, "CML GT1", "Intel(R) UHD Graphics")
CHIPSET(0x9ba5, cfl_gt1, "CML GT1", "Intel(R) UHD Graphics 610")
CHIPSET(0x9ba8, cfl_gt1, "CML GT1", "Intel(R) UHD Graphics 610")
CHIPSET(0x9baa, cfl_gt1, "CML GT1", "Intel(R) UHD Graphics")
CHIPSET(0x9bab, cfl_gt1, "CML GT1", "Intel(R) UHD Graphics")
CHIPSET(0x9bac, cfl_gt1, "CML GT1", "Intel(R) UHD Graphics")
CHIPSET(0x9b41, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9bc0, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9bc2, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9bc4, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9bc5, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics 630")
CHIPSET(0x9bc6, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics P630")
CHIPSET(0x9bc8, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics 630")
CHIPSET(0x9bca, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9bcb, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9bcc, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9be6, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics P630")
CHIPSET(0x9bf6, cfl_gt2, "CML GT2", "Intel(R) UHD Graphics P630")

CHIPSET(0x8a50, icl_gt2,   "ICL GT2",   "Intel(R) HD Graphics")
CHIPSET(0x8a51, icl_gt2,   "ICL GT2",   "Intel(R) Iris(R) Plus Graphics")
CHIPSET(0x8a52, icl_gt2,   "ICL GT2",   "Intel(R) Iris(R) Plus Graphics")
CHIPSET(0x8a53, icl_gt2,   "ICL GT2",   "Intel(R) Iris(R) Plus Graphics")
CHIPSET(0x8a54, icl_gt1_5, "ICL GT1.5", "Intel(R) Iris(R) Plus Graphics")
CHIPSET(0x8a56, icl_gt1,   "ICL GT1",   "Intel(R) UHD Graphics")
CHIPSET(0x8a57, icl_gt1_5, "ICL GT1.5", "Intel(R) HD Graphics")
CHIPSET(0x8a58, icl_gt1,   "ICL GT1",   "Intel(R) UHD Graphics")
CHIPSET(0x8a59, icl_gt1_5, "ICL GT1.5", "Intel(R) HD Graphics")
CHIPSET(0x8a5a, icl_gt1_5, "ICL GT1.5", "Intel(R) Iris(R) Plus Graphics")
CHIPSET(0x8a5b, icl_gt1,   "ICL GT1",   "Intel(R) HD Graphics")
CHIPSET(0x8a5c, icl_gt1_5, "ICL GT1.5", "Intel(R) Iris(R) Plus Graphics")
CHIPSET(0x8a5d, icl_gt1,   "ICL GT1",   "Intel(R) HD Graphics")
CHIPSET(0x8a71, icl_gt0_5, "ICL GT0.5", "Intel(R) HD Graphics")

CHIPSET(0x4541, ehl_2x4, "EHL", "Intel(R) UHD Graphics")
CHIPSET(0x4551, ehl_4x4, "EHL", "Intel(R) UHD Graphics")
CHIPSET(0x4555, ehl_2x8, "EHL", "Intel(R) UHD Graphics")
CHIPSET(0x4557, ehl_4x5, "EHL", "Intel(R) UHD Graphics")
CHIPSET(0x4570, ehl_4x8, "EHL", "Intel(R) UHD Graphics")
CHIPSET(0x4571, ehl_4x8, "EHL", "Intel(R) UHD Graphics")
CHIPSET(0x4e51, ehl_4x4, "JSL", "Intel(R) UHD Graphics")
CHIPSET(0x4e55, ehl_2x8, "JSL", "Intel(R) UHD Graphics")
CHIPSET(0x4e57, ehl_4x5, "JSL", "Intel(R) UHD Graphics")
CHIPSET(0x4e61, ehl_4x6, "JSL", "Intel(R) UHD Graphics")
CHIPSET(0x4e71, ehl_4x8, "JSL", "Intel(R) UHD Graphics")

CHIPSET(0x4c8a, rkl_gt1, "RKL GT1", "Intel(R) Graphics")
CHIPSET(0x4c8b, rkl_gt1, "RKL GT1", "Intel(R) Graphics")
CHIPSET(0x4c8c, rkl_gt05, "RKL GT0.5", "Intel(R) Graphics")
CHIPSET(0x4c90, rkl_gt1, "RKL GT1", "Intel(R) Graphics")
CHIPSET(0x4c9a, rkl_gt1, "RKL GT1", "Intel(R) Graphics")

CHIPSET(0x4680, adl_gt1, "ADL-S GT1", "Intel(R) UHD Graphics 770")
CHIPSET(0x4682, adl_gt1, "ADL-S GT1", "Intel(R) UHD Graphics 730")
CHIPSET(0x4688, adl_gt1, "ADL-S GT1", "Intel(R) UHD Graphics")
CHIPSET(0x468a, adl_gt1, "ADL-S GT1", "Intel(R) UHD Graphics")
CHIPSET(0x468b, adl_gt05, "ADL-S GT0.5", "Intel(R) UHD Graphics")
CHIPSET(0x4690, adl_gt1, "ADL-S GT1", "Intel(R) UHD Graphics 770")
CHIPSET(0x4692, adl_gt1, "ADL-S GT1", "Intel(R) UHD Graphics 730")
CHIPSET(0x4693, adl_gt1, "ADL-S GT1", "Intel(R) UHD Graphics 710")

CHIPSET(0x4626, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x4628, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x462a, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46a0, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46a1, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46a2, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46a3, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46a6, adl_gt2, "ADL GT2", "Intel(R) Iris(R) Xe Graphics")
CHIPSET(0x46a8, adl_gt2, "ADL GT2", "Intel(R) Iris(R) Xe Graphics")
CHIPSET(0x46aa, adl_gt2, "ADL GT2", "Intel(R) Iris(R) Xe Graphics")
CHIPSET(0x46b0, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46b1, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46b2, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46b3, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46c0, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46c1, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46c2, adl_gt2, "ADL GT2", "Intel(R) Graphics")
CHIPSET(0x46c3, adl_gt2, "ADL GT2", "Intel(R) Graphics")

CHIPSET(0x46d0, adl_gt1,  "ADL-N", "Intel(R) Graphics")
CHIPSET(0x46d1, adl_gt1,  "ADL-N", "Intel(R) Graphics")
CHIPSET(0x46d2, adl_gt05, "ADL-N", "Intel(R) Graphics")
CHIPSET(0x46d3, adl_gt1,  "ADL-N", "Intel(R) Graphics")
CHIPSET(0x46d4, adl_gt1,  "ADL-N", "Intel(R) Graphics")

CHIPSET(0x9a40, tgl_gt2, "TGL GT2", "Intel(R) Iris(R) Xe Graphics")
CHIPSET(0x9a49, tgl_gt2, "TGL GT2", "Intel(R) Iris(R) Xe Graphics")
CHIPSET(0x9a59, tgl_gt2, "TGL GT2", "Intel(R) Graphics")
CHIPSET(0x9a60, tgl_gt1, "TGL GT1", "Intel(R) UHD Graphics")
CHIPSET(0x9a68, tgl_gt1, "TGL GT1", "Intel(R) UHD Graphics")
CHIPSET(0x9a70, tgl_gt1, "TGL GT1", "Intel(R) UHD Graphics")
CHIPSET(0x9a78, tgl_gt2, "TGL GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9ac0, tgl_gt2, "TGL GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9ac9, tgl_gt2, "TGL GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9ad9, tgl_gt2, "TGL GT2", "Intel(R) UHD Graphics")
CHIPSET(0x9af8, tgl_gt2, "TGL GT2", "Intel(R) UHD Graphics")

CHIPSET(0xa780, rpl, "RPL-S", "Intel(R) Graphics")
CHIPSET(0xa781, rpl, "RPL-S", "Intel(R) Graphics")
CHIPSET(0xa782, rpl, "RPL-S", "Intel(R) Graphics")
CHIPSET(0xa783, rpl, "RPL-S", "Intel(R) Graphics")
CHIPSET(0xa788, rpl, "RPL-S", "Intel(R) Graphics")
CHIPSET(0xa789, rpl, "RPL-S", "Intel(R) Graphics")
CHIPSET(0xa78a, rpl, "RPL-S", "Intel(R) Graphics")
CHIPSET(0xa78b, rpl, "RPL-S", "Intel(R) Graphics")

CHIPSET(0xa720, rpl_p, "RPL-P", "Intel(R) Graphics")
CHIPSET(0xa721, rpl_p, "RPL-U", "Intel(R) Graphics")
CHIPSET(0xa7a0, rpl_p, "RPL-P", "Intel(R) Iris(R) Xe Graphics")
CHIPSET(0xa7a1, rpl_p, "RPL-U", "Intel(R) Iris(R) Xe Graphics")
CHIPSET(0xa7a8, rpl_p, "RPL-P", "Intel(R) Graphics")
CHIPSET(0xa7a9, rpl_p, "RPL-U", "Intel(R) Graphics")
CHIPSET(0xa7aa, rpl_p, "RPL-P", "Intel(R) Graphics")
CHIPSET(0xa7ab, rpl_p, "RPL-P", "Intel(R) Graphics")
CHIPSET(0xa7ac, rpl_p, "RPL-U", "Intel(R) Graphics")
CHIPSET(0xa7ad, rpl_p, "RPL-U", "Intel(R) Graphics")

CHIPSET(0x4905, dg1, "DG1", "Intel(R) Iris(R) Xe MAX Graphics")
CHIPSET(0x4906, dg1, "DG1", "Intel(R) Graphics")
CHIPSET(0x4907, sg1, "SG1", "Intel(R) Graphics")
CHIPSET(0x4908, dg1, "DG1", "Intel(R) Iris(R) Xe Graphics")
CHIPSET(0x4909, dg1, "DG1", "Intel(R) Graphics")

CHIPSET(0x5690, dg2_g10, "DG2", "Intel(R) Arc(tm) A770M Graphics")
CHIPSET(0x5691, dg2_g10, "DG2", "Intel(R) Arc(tm) A730M Graphics")
CHIPSET(0x5692, dg2_g10, "DG2", "Intel(R) Arc(tm) A550M Graphics")
CHIPSET(0x5693, dg2_g11, "DG2", "Intel(R) Arc(tm) A370M Graphics")
CHIPSET(0x5694, dg2_g11, "DG2", "Intel(R) Arc(tm) A350M Graphics")
CHIPSET(0x5695, dg2_g11, "DG2", "Intel(R) Graphics")
CHIPSET(0x5696, dg2_g12, "DG2", "Intel(R) Arc(tm) A570M Graphics")
CHIPSET(0x5697, dg2_g12, "DG2", "Intel(R) Arc(tm) A530M Graphics")
CHIPSET(0x56a0, dg2_g10, "DG2", "Intel(R) Arc(tm) A770 Graphics")
CHIPSET(0x56a1, dg2_g10, "DG2", "Intel(R) Arc(tm) A750 Graphics")
CHIPSET(0x56a2, dg2_g10, "DG2", "Intel(R) Arc(tm) A580 Graphics")
CHIPSET(0x56a3, dg2_g12, "DG2", "Intel(R) Graphics")
CHIPSET(0x56a4, dg2_g12, "DG2", "Intel(R) Graphics")
CHIPSET(0x56a5, dg2_g11, "DG2", "Intel(R) Arc(tm) A380 Graphics")
CHIPSET(0x56a6, dg2_g11, "DG2", "Intel(R) Arc(tm) A310 Graphics")
CHIPSET(0x56b0, dg2_g11, "DG2", "Intel(R) Arc(tm) Pro A30M Graphics")
CHIPSET(0x56b1, dg2_g11, "DG2", "Intel(R) Arc(tm) Pro A40/A50 Graphics")
CHIPSET(0x56b2, dg2_g12, "DG2", "Intel(R) Arc(tm) Pro A60M Graphics")
CHIPSET(0x56b3, dg2_g12, "DG2", "Intel(R) Arc(tm) Pro A60 Graphics")
CHIPSET(0x56ba, dg2_g11, "DG2", "Intel(R) Arc(tm) A380E Graphics")
CHIPSET(0x56bb, dg2_g11, "DG2", "Intel(R) Arc(tm) A310E Graphics")
CHIPSET(0x56bc, dg2_g11, "DG2", "Intel(R) Arc(tm) A370E Graphics")
CHIPSET(0x56bd, dg2_g11, "DG2", "Intel(R) Arc(tm) A350E Graphics")
CHIPSET(0x56be, dg2_g10, "DG2", "Intel(R) Arc(tm) A750E Graphics")
CHIPSET(0x56bf, dg2_g10, "DG2", "Intel(R) Arc(tm) A580E Graphics")
CHIPSET(0x56c0, atsm_g10, "ATS-M", "Intel(R) Data Center GPU Flex 170")
CHIPSET(0x56c1, atsm_g11, "ATS-M", "Intel(R) Data Center GPU Flex 140")
CHIPSET(0x56c2, atsm_g10, "ATS-M", "Intel(R) Data Center GPU Flex 170V")

CHIPSET(0x6420, lnl, "LNL", "Intel(R) Graphics")
CHIPSET(0x64a0, lnl, "LNL", "Intel(R) Graphics")
CHIPSET(0x64b0, lnl, "LNL", "Intel(R) Graphics")

CHIPSET(0x7d40, mtl_u, "MTL", "Intel(R) Graphics")
CHIPSET(0x7d45, mtl_u, "MTL", "Intel(R) Graphics")
CHIPSET(0x7d55, mtl_h, "MTL", "Intel(R) Arc(tm) Graphics")
CHIPSET(0x7d60, mtl_u, "MTL", "Intel(R) Graphics")
CHIPSET(0x7dd5, mtl_h, "MTL", "Intel(R) Graphics")

CHIPSET(0x7d41, arl_u, "ARL", "Intel(R) Graphics")
CHIPSET(0x7d51, arl_h, "ARL", "Intel(R) Graphics")
CHIPSET(0x7d67, arl_u, "ARL", "Intel(R) Graphics")
CHIPSET(0x7dd1, arl_h, "ARL", "Intel(R) Graphics")
CHIPSET(0xb640, arl_u, "ARL", "Intel(R) Graphics")

CHIPSET(0xe202, bmg, "BMG G21", "Intel(R) Graphics")
CHIPSET(0xe209, bmg, "BMG G21", "Intel(R) Graphics")
CHIPSET(0xe20b, bmg, "BMG G21", "Intel(R) Arc(tm) B580 Graphics")
CHIPSET(0xe20c, bmg, "BMG G21", "Intel(R) Arc(tm) B570 Graphics")
CHIPSET(0xe20d, bmg, "BMG G21", "Intel(R) Graphics")
CHIPSET(0xe210, bmg, "BMG G21", "Intel(R) Graphics")
CHIPSET(0xe211, bmg, "BMG G21", "Intel(R) Arc(tm) Pro B60 Graphics")
CHIPSET(0xe212, bmg, "BMG G21", "Intel(R) Arc(tm) Pro B50 Graphics")
CHIPSET(0xe215, bmg, "BMG", "Intel(R) Graphics")
CHIPSET(0xe216, bmg, "BMG", "Intel(R) Graphics")
CHIPSET(0xe220, bmg, "BMG G31", "Intel(R) Graphics")
CHIPSET(0xe221, bmg, "BMG G31", "Intel(R) Graphics")
CHIPSET(0xe222, bmg, "BMG G31", "Intel(R) Graphics")
CHIPSET(0xe223, bmg, "BMG G31", "Intel(R) Graphics")

CHIPSET(0xb080, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb081, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb082, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb083, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb084, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb085, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb086, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb087, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb08f, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb090, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb0a0, ptl, "PTL", "Intel(R) Graphics")
CHIPSET(0xb0b0, ptl, "PTL", "Intel(R) Graphics")

CHIPSET(0xfd80, wcl, "WCL", "Intel(R) Graphics")
CHIPSET(0xfd81, wcl, "WCL", "Intel(R) Graphics")

CHIPSET(0xd740, nvl_s_hx_ul, "NVL-S", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd741, nvl_u_h, "NVL-U", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd742, nvl_u_h, "NVL-H", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd743, nvl_s_hx_ul, "NVL-HX", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd744, nvl_s_hx_ul, "NVL-UL", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd745, nvl_s_hx_ul, "NVL-HX", "Intel(R) Graphics", FORCE_PROBE)

CHIPSET(0xd750, nvl_p, "NVL-P", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd751, nvl_p, "NVL-P", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd752, nvl_p, "NVL-P", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd753, nvl_p, "NVL-P", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd754, nvl_p, "NVL-P", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd755, nvl_p, "NVL-P", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd756, nvl_p, "NVL-P", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd757, nvl_p, "NVL-P", "Intel(R) Graphics", FORCE_PROBE)
CHIPSET(0xd75f, nvl_p, "NVL-P", "Intel(R) Graphics", FORCE_PROBE)
